use nom::branch::alt;
use nom::character::complete::char;
use nom::IResult;

#[derive(Copy, Clone, Ord, PartialOrd, Eq, PartialEq, Debug)]
pub struct Trick {
    thrown: Hand,
    against: Hand,
}

impl From<(Hand, Hand)> for Trick {
    fn from(value: (Hand, Hand)) -> Self {
        Self {
            thrown: value.1,
            against: value.0,
        }
    }
}

#[derive(Copy, Clone, Ord, PartialOrd, Eq, PartialEq, Debug)]
enum Hand {
    Rock,
    Paper,
    Scissors,
}

const WIN: usize = 6;
const TIE: usize = 3;
const LOSS: usize = 0;

impl Trick {
    fn score(&self) -> usize {
        let mut score = 0;
        match &self.thrown {
            Hand::Rock => score += 1,
            Hand::Paper => score += 2,
            Hand::Scissors => score += 3,
        };
        match (&self.thrown, &self.against) {
            (Hand::Rock, Hand::Rock) => score += TIE,
            (Hand::Rock, Hand::Paper) => score += LOSS,
            (Hand::Rock, Hand::Scissors) => score += WIN,
            (Hand::Paper, Hand::Rock) => score += WIN,
            (Hand::Paper, Hand::Paper) => score += TIE,
            (Hand::Paper, Hand::Scissors) => score += LOSS,
            (Hand::Scissors, Hand::Rock) => score += LOSS,
            (Hand::Scissors, Hand::Paper) => score += WIN,
            (Hand::Scissors, Hand::Scissors) => score += TIE,
        }
        score
    }
}

fn rock_against(s: &str) -> IResult<&str, Hand> {
    let (rest, _) = char('A')(s)?;
    Ok((rest, Hand::Rock))
}

fn paper_against(s: &str) -> IResult<&str, Hand> {
    let (rest, _) = char('B')(s)?;
    Ok((rest, Hand::Paper))
}
fn scissors_against(s: &str) -> IResult<&str, Hand> {
    let (rest, _) = char('C')(s)?;
    Ok((rest, Hand::Scissors))
}

fn against_hand_parser(s: &str) -> IResult<&str, Hand> {
    alt((rock_against, paper_against, scissors_against))(s)
}

#[derive(Copy, Clone, Ord, PartialOrd, Eq, PartialEq, Debug)]
enum ResultExpected {
    Win,
    Lose,
    Draw,
}

fn loss_expected(s: &str) -> IResult<&str, ResultExpected> {
    let (rest, _) = char('X')(s)?;
    Ok((rest, ResultExpected::Lose))
}
fn draw_expected(s: &str) -> IResult<&str, ResultExpected> {
    let (rest, _) = char('Y')(s)?;
    Ok((rest, ResultExpected::Draw))
}
fn win_expected(s: &str) -> IResult<&str, ResultExpected> {
    let (rest, _) = char('Z')(s)?;
    Ok((rest, ResultExpected::Win))
}

fn result_expected_parser(s: &str) -> IResult<&str, ResultExpected> {
    alt((loss_expected, draw_expected, win_expected))(s)
}

impl From<(Hand, ResultExpected)> for Trick {
    fn from(value: (Hand, ResultExpected)) -> Self {
        match value.1 {
            ResultExpected::Win => match value.0 {
                against @ Hand::Rock => Trick {
                    against,
                    thrown: Hand::Paper,
                },
                against @ Hand::Paper => Trick {
                    against,
                    thrown: Hand::Scissors,
                },
                against @ Hand::Scissors => Trick {
                    against,
                    thrown: Hand::Rock,
                },
            },
            ResultExpected::Lose => match value.0 {
                against @ Hand::Rock => Trick {
                    against,
                    thrown: Hand::Scissors,
                },
                against @ Hand::Paper => Trick {
                    against,
                    thrown: Hand::Rock,
                },
                against @ Hand::Scissors => Trick {
                    against,
                    thrown: Hand::Paper,
                },
            },
            ResultExpected::Draw => match value.0 {
                against @ Hand::Rock => Trick {
                    against,
                    thrown: against,
                },
                against @ Hand::Paper => Trick {
                    against,
                    thrown: against,
                },
                against @ Hand::Scissors => Trick {
                    against,
                    thrown: against,
                },
            },
        }
    }
}

pub(crate) mod part_02;

pub(crate) mod part_01;
