use crate::parser::{
    against_hand_parser, paper_against, rock_against, scissors_against, Hand, Trick,
};
use nom::branch::alt;
use nom::character::complete::{char, line_ending};
use nom::multi::separated_list1;
use nom::sequence::separated_pair;
use nom::IResult;

fn rock_thrown(s: &str) -> IResult<&str, Hand> {
    let (rest, _) = char('X')(s)?;
    Ok((rest, Hand::Rock))
}

fn paper_thrown(s: &str) -> IResult<&str, Hand> {
    let (rest, _) = char('Y')(s)?;
    Ok((rest, Hand::Paper))
}

fn scissors_thrown(s: &str) -> IResult<&str, Hand> {
    let (rest, _) = char('Z')(s)?;
    Ok((rest, Hand::Scissors))
}

fn thrown_hand_parser(s: &str) -> IResult<&str, Hand> {
    alt((rock_thrown, paper_thrown, scissors_thrown))(s)
}

fn trick_parser(s: &str) -> IResult<&str, Trick> {
    let (rest, hands): (&str, (Hand, Hand)) =
        separated_pair(against_hand_parser, char(' '), thrown_hand_parser)(s)?;
    Ok((rest, hands.into()))
}

fn sequence_parser(s: &str) -> IResult<&str, Vec<Trick>> {
    separated_list1(line_ending, trick_parser)(s)
}

pub(crate) fn parse(inp: &str) -> Vec<usize> {
    let (_, tricks): (&str, Vec<Trick>) = sequence_parser(inp).expect("Couldn't parse input");
    tricks.iter().map(|trick| trick.score()).collect()
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn against_rock_parsed() {
        assert_eq!(Ok(("", Hand::Rock)), rock_against("A"))
    }
    #[test]
    fn thrown_rock_parsed() {
        assert_eq!(Ok(("", Hand::Rock)), rock_thrown("X"))
    }

    // Idea stolen from https://stackoverflow.com/a/34666891
    macro_rules! param_test {
        ($name:ident, $input:literal, $against:expr, $thrown:expr) => {
            #[test]
            fn $name() {
                assert_eq!(
                    Ok((
                        "",
                        Trick {
                            thrown: $thrown,
                            against: $against
                        }
                    )),
                    trick_parser($input)
                )
            }
        };
    }

    param_test!(rock_against_rock_thrown, "A X", Hand::Rock, Hand::Rock);
    param_test!(rock_against_paper_thrown, "A Y", Hand::Rock, Hand::Paper);
    param_test!(
        rock_against_scissors_thrown,
        "A Z",
        Hand::Rock,
        Hand::Scissors
    );

    param_test!(paper_against_rock_thrown, "B X", Hand::Paper, Hand::Rock);
    param_test!(paper_against_paper_thrown, "B Y", Hand::Paper, Hand::Paper);
    param_test!(
        paper_against_scissors_thrown,
        "B Z",
        Hand::Paper,
        Hand::Scissors
    );

    param_test!(
        scissors_against_rock_thrown,
        "C X",
        Hand::Scissors,
        Hand::Rock
    );
    param_test!(
        scissors_against_paper_thrown,
        "C Y",
        Hand::Scissors,
        Hand::Paper
    );
    param_test!(
        scissors_against_scissors_thrown,
        "C Z",
        Hand::Scissors,
        Hand::Scissors
    );

    #[test]
    fn example_sequence_parsed() {
        assert_eq!(
            Ok((
                "",
                vec![
                    Trick {
                        thrown: Hand::Paper,
                        against: Hand::Rock
                    },
                    Trick {
                        against: Hand::Paper,
                        thrown: Hand::Rock
                    },
                    Trick {
                        against: Hand::Scissors,
                        thrown: Hand::Scissors
                    }
                ]
            )),
            sequence_parser("A Y\nB X\nC Z")
        )
    }
}
